
#include <xccompat.h>
#include <hwtimer.h>
#include <stdint.h>
#include <xs2a_user.h>
#include <xs1.h>
#include <debug_print.h>
#include "pll_timer.h"

#define ISR_ENTER(n) \
    asm (".text"); \
    asm (".globl " #n); \
    asm (#n ":"); \
    asm ("kentsp 22"); \
    asm ("stw spc, sp[1]"); \
    asm ("stw ssr, sp[2]"); \
    asm ("stw sed, sp[3]"); \
    asm ("stw et, sp[4]"); \
    asm ("stw dp, sp[5]"); \
    asm ("stw cp, sp[6]"); \
    asm ("stw lr, sp[7]"); \
    asm ("stw r0, sp[8]"); \
    asm ("stw r1, sp[9]"); \
    asm ("stw r2, sp[10]"); \
    asm ("stw r3, sp[11]"); \
    asm ("stw r4, sp[12]"); \
    asm ("stw r5, sp[13]"); \
    asm ("stw r6, sp[14]"); \
    asm ("stw r7, sp[15]"); \
    asm ("stw r8, sp[16]"); \
    asm ("stw r9, sp[17]"); \
    asm ("stw r10, sp[18]"); \
    asm ("stw r11, sp[19]");

#define ISR_EXIT \
    asm ("ldw spc, sp[1]"); \
    asm ("ldw ssr, sp[2]"); \
    asm ("ldw sed, sp[3]"); \
    asm ("ldw et, sp[4]"); \
    asm ("ldw dp, sp[5]"); \
    asm ("ldw cp, sp[6]"); \
    asm ("ldw lr, sp[7]"); \
    asm ("ldw r0, sp[8]"); \
    asm ("ldw r1, sp[9]"); \
    asm ("ldw r2, sp[10]"); \
    asm ("ldw r3, sp[11]"); \
    asm ("ldw r4, sp[12]"); \
    asm ("ldw r5, sp[13]"); \
    asm ("ldw r6, sp[14]"); \
    asm ("ldw r7, sp[15]"); \
    asm ("ldw r8, sp[16]"); \
    asm ("ldw r9, sp[17]"); \
    asm ("ldw r10, sp[18]"); \
    asm ("ldw r11, sp[19]"); \
    asm ("krestsp 22"); \
    asm ("kret");

#define ISR_DECL(n, c, a) \
    ISR_ENTER(n) \
    asm (".globl " #c); \
    asm ("ldw r0, sp[21]"); \
    asm ("bl " #c); \
    ISR_EXIT 

void __pll_ctrl_int(void);
ISR_DECL(__pll_ctrl_int, pll_ctrl_int, 0)


#define TIMER_PERIOD   100000000/10000

static hwtimer_t ctrl_timer;
static int32_t tmr_count = 0;
static int32_t last_int_time = 0;
static uint32_t int_counter = 0;

/*
 * handler for pll timer interrupt
 */
void pll_ctrl_int(void *arg)
{
    ++int_counter;
    tmr_count += 100000000/11111;
    asm volatile("setd res[%0], %1"::"r"(ctrl_timer), "r"(tmr_count));
    asm volatile("gettime %0":"=r"(last_int_time));
}


/*
 * set up timer interrupt for pll control 
 */
#define STACK_SZ64 32
#define STACK_SZ32 (STACK_SZ64 * 2)
uint64_t isr_stack[STACK_SZ64] = {[0 ... STACK_SZ64 - 1] = 0xa5a5a5a5a5a5a5a5};

void pll_ctrl_init(void)
{
    // set up interrupt stack
    uint32_t cur_sp;
        // save current sp
    asm volatile("ldaw %0, sp[0]" : "=r"(cur_sp));
        // set current sp to end of int stack (64-bit align)
    asm volatile("set sp, %0" : : "r"(&isr_stack[STACK_SZ64 - 1]));
        // store current sp where krestsp expects to find it
    asm volatile("stw %0, sp[0]" : : "r"(cur_sp));
        // krestsp (sp -> ksp, sp[0] -> sp)
    asm volatile("krestsp 0");
    
    // grab a hardware timer
    asm volatile("getr %0, %1" : "=r"(ctrl_timer) : "i"(XS1_RES_TYPE_TIMER));
    
    // interrupt mode
    asm volatile("setv res[%0], %1" : : "r"(ctrl_timer), "r"(__pll_ctrl_int));
    asm volatile("setc res[%0], %1" : : "r"(ctrl_timer), "i"(XS1_SETC_IE_MODE_INTERRUPT));
    asm volatile("setc res[%0], %1" : : "r"(ctrl_timer), "i"(XS1_SETC_COND_AFTER));
    
    // grab current time and set time for first interrupt
    asm volatile("in %0, res[%1]" : "=r"(tmr_count) : "r"(ctrl_timer));
    tmr_count += TIMER_PERIOD;
    asm volatile("setd res[%0], %1" : : "r"(ctrl_timer), "r"(tmr_count));
    
    // enable interrupts/events from timer and unmask interrupts
    asm volatile("eeu res[%0]" : : "r"(ctrl_timer));
    asm volatile("setsr 2");
}


void pll_ctrl_get_stats(uint32_t *count, int32_t *stamp)
{
    asm volatile("clrsr 2");
    *count = int_counter;
    *stamp = last_int_time;
    asm volatile("setsr 2");
}

